/* ------------------------------------------------------------------------ */
/*                                                                          */
/*  Copyright (C)                                                           */
/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 2 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful, but     */
/*  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANT-    */
/*  ABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General        */
/*  Public License for more details.                                        */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA                   */
/*                                                                          */
/* ------------------------------------------------------------------------ */
/*  $Author: vogje02 $ $Revision: 1.7 $ $Date: 2002/07/28 17:13:19 $        */
/* ------------------------------------------------------------------------ */

#include <nmapserv.h>

/**
 * Signal handler. This is geeting the SIGINT notifications (Crtl-C) and stops the 
 * the foreground process.
 */
void SignalHandler( int s )
{
  switch( s ) {

  case SIGINT:
    fprintf(stderr, "Got SIGINT signal\n" );
    fprintf(stderr, "Aborting...\n" );    
    exit(1);
    break;
  }
}

/**
 * Print the help screen.
 */
void usage()
{
  fprintf(stderr, "\nnmapserv v%s\n", version );
  fprintf(stderr, "Copyright (c) 2000 - 2002 JV Software (vogje01@bluewin.ch)\n\n" );
  fprintf(stderr, "usage:\n");
  fprintf(stderr, "\tnmapserv [-i|-u|-p [-d n]] [-h]\n\n");
  fprintf(stderr, "Options:\n");
  fprintf(stderr, "\t-h\t\t: This help screen.\n");
  fprintf(stderr, "\t-i\t\t: Install service.\n");
  fprintf(stderr, "\t-u\t\t: Uninstall service.\n");
  fprintf(stderr, "\t-p\t\t: Run as foreground process.\n");
  fprintf(stderr, "\t-d n\t\t: Debug level (only with -p).\n\n");
}

/**
 * Constructor.
 */
CNMapService::CNMapService( CLogWriter *lw ) : CNTService( "NMap", "NMap", lw )
{
	// Service initialization
	m_iStartParam = 0;
	m_iIncParam = 1;
	m_iState = m_iStartParam;

	// Make local copies
	logWriter = lw;

	// Default arguments
	sprintf( (char*)nmapArgs, "-sS -PI -PT" );

	// Auto start flag
	autoStart = false;
}

/**
 * Win32 service initialization
 */
BOOL CNMapService::OnInit()
{
	HKEY hkey;
	char szKey[1024];

	logWriter->log( INFO, "CNMapService", "OnInit", "Starting - service: %s", m_szServiceName );

	// Read the registry parameters. Try opening the registry key:
	// HKEY_LOCAL_MACHINE\SYSTEM\CurrentControlSet\Services\nmapserv\Parameters
	strcpy(szKey, "SYSTEM\\CurrentControlSet\\Services\\NMap\\Parameters");

	if( RegOpenKeyEx(HKEY_LOCAL_MACHINE, szKey, 0, KEY_QUERY_VALUE|KEY_NOTIFY, &hkey ) == ERROR_SUCCESS ) {

		// Yes we are installed
		DWORD dwType = 0;
		DWORD dwSize = sizeof( m_iStartParam );
		RegQueryValueEx( hkey, "Start", NULL, &dwType, (BYTE*)&m_iStartParam, &dwSize );

		dwSize = sizeof( m_iIncParam );
		RegQueryValueEx( hkey, "Inc", NULL, &dwType, (BYTE*)&m_iIncParam, &dwSize );

		RegCloseKey(hkey);
	}

	// Set the initial state
	m_iState = m_iStartParam;

	// Read the nmap and nmapserv arguments from
	// the registry
	if( ReadRegistry() ) {

		// Parse the nmapserv arguments
		ParseServiceArgs();

		// Get the next scanning time.
		GetNextScan();
	}

	return TRUE;
}

/**
 * Process user control requests
 */ 
BOOL CNMapService::OnUserControl( DWORD dwOpcode )
{
  switch (dwOpcode) {
    
  case SERVICE_CONTROL_USER + 0:
    
    // Save the current status in the registry
    SaveStatus();
    return TRUE;
    
  default:
    break;
  }
  return FALSE; // say not handled
}

/**
 * Save the current status in the registry
 */
void CNMapService::SaveStatus()
{
  logWriter->log( INFO, "CNMapService", "SaveStatus", "Saving current status" );
  
  // Try opening the registry key:
  // HKEY_LOCAL_MACHINE\SYSTEM\CurrentControlSet\Services\<AppName>\...
  HKEY hkey = NULL;
  char szKey[1024];
  strcpy(szKey, "SYSTEM\\CurrentControlSet\\Services\\");
  strcat(szKey, m_szServiceName);
  strcat(szKey, "\\Status");
  DWORD dwDisp;
  DWORD dwErr;
  
  logWriter->log( DEBUG0, "CNMapService", "SaveStatus", "Creating key: %s", szKey );
  
  dwErr = RegCreateKeyEx( HKEY_LOCAL_MACHINE, szKey, 0, "", REG_OPTION_NON_VOLATILE, KEY_WRITE, NULL, &hkey, &dwDisp );
  if (dwErr != ERROR_SUCCESS) {
    logWriter->log( ERROR, "CNMapService", "SaveStatus", "Failed to create Status key (%lu)", dwErr );
    return;
  }	
  
  // Set the registry values
  logWriter->log( DEBUG0, "CNMapService", "SaveStatus", "Saving 'Current' as %ld", m_iState );
  RegSetValueEx( hkey, "Current", 0, REG_DWORD, (BYTE*)&m_iState, sizeof( m_iState ) );

  // Finished with key
  RegCloseKey(hkey);
}

/**
 * Stops a running nmap process on service shutdown
 */
void CNMapService::OnStop()
{
	int pid = -1, nThreads = -1;

    // If we have a process pid > 0
    if( ( pid = GetPid( "nmap", &nThreads ) ) > 0 ) {
		logWriter->log( DEBUG0, "CNMapService", "OnStop", "Stopping nmap service PID: %d", pid );
		TerminateProcess( pi.hProcess, 0 );
	}
}

/**
 * This routine is starting the nmap process after each interval. It also checks the
 * 'registry change' events and start the reread of the nmap arguments.
 */
void CNMapService::RunService()
{ 
	HKEY hkey;
	char szKey[1024];
	int ret;

	// Read the registry parameters. Try opening the registry key:
	// HKEY_LOCAL_MACHINE\SYSTEM\CurrentControlSet\Services\nmapserv\Parameters
	strcpy(szKey, "SYSTEM\\CurrentControlSet\\Services\\NMap\\Parameters");
	if( RegOpenKeyEx(HKEY_LOCAL_MACHINE, szKey, 0, KEY_NOTIFY, &hkey ) != ERROR_SUCCESS ) {
		logWriter->log( ERROR, "CNMapService", "RunService", "Could not open registry key '%s'",
						szKey );
		exit(1);
	}

	// Create the event object. The NT kernel signals this event when 
	// it receives the 'change registry value' control code. 
	if( ( hEvent = CreateEvent( NULL, FALSE, FALSE, NULL ) ) == NULL )
		logWriter->log( ERROR, "CNMapService", "RunService", "Could not create event handle" );

	if( RegNotifyChangeKeyValue( hkey, FALSE, REG_NOTIFY_CHANGE_LAST_SET, hEvent, TRUE ) != ERROR_SUCCESS ) {
		logWriter->log( ERROR, "CNMapService", "RunService", "Could not create event handle" );
	
	}

	// Start scanner, perform one scan during servive startup, if autostart == true. 
	// Next scan will be startTime+factor*interval.
	if( autoStart && immediately )
		StartScanner();
	else
		logWriter->log( INFO, "CNMapService", "RunService", "AutoStart: OFF" );
		
	// Get the next scanning time.
	GetNextScan();

	// Start endless loop
	logWriter->log( INFO, "CNMapService", "RunService", "Starting processing loop: %d", m_bIsRunning );
	while( m_bIsRunning ) {

		// Check current time and start scanner if necessary
		long ltime;
		time( &ltime );
		CTime now = CTime( ltime );
		logWriter->log( DEBUG1, "CNMapService", "RunService", "Current time: %d (%d) nextScan: %d", now.GetTime(), days[now.GetDayOfWeek()], nextScan );

		if( ( now.GetTime() > nextScan ) && days[now.GetDayOfWeek()-1] ) {
	
			// Start scanner, if autostart == true
			if( autoStart )
				StartScanner();

			// Get the next scanning time.
			GetNextScan();
		}

		// Wait for the registry change event for 1 sec. This is also used
		// as the general timer.
		if( WaitForSingleObject( hEvent, 1000 ) == WAIT_OBJECT_0 ) {

			logWriter->log( INFO, "CNMapService", "RunService", "Got 'registry changed' event" );

			// Get the new arguments for nmap and nmapserv.
			if( ReadRegistry() ) {

				// Parse the nmapserv arguments
				ParseServiceArgs();

				// Get the next scanning time.
				GetNextScan();

				// If immediately is checked
				if( immediately )
					StartScanner();
			}

			// Restart the notification listener
			if( ( ret = RegNotifyChangeKeyValue( hkey, FALSE, REG_NOTIFY_CHANGE_LAST_SET, 
												 hEvent, TRUE ) ) != ERROR_SUCCESS )
				logWriter->log( INFO, "CNMapService", "RunService", 
								"Could not install notification event" );
				
		}
	}
}

/**
 * Get the process pid by name
 */
int CNMapService::GetPid( char* appl, int* nThreads )
{
  int pid;
  char						name[80];
  PPERF_OBJECT		nObject;
  PPERF_INSTANCE	nInstance;	
  
  /*
   * Initilize performance data
   */
  SetPerfIndexes();
  
  /*
   * Cleanup
   */
  LocalFree(gpPerfData);
  gpPerfData = NULL;
  
  /* 
   * Get the performance data
   */
  if(GetPerfData (hPerfKey, INDEX_PROCTHRD_OBJ, &gpPerfData, &gpPerfDataSize) != 0)
    fprintf(stderr, "Could not get PerfData" );
  
  /*
   * Set first object/instance
   */
  nObject = FindObject (gpPerfData, PX_PROCESS);
  nInstance = FirstInstance(nObject);		
  
  /*
   * Loop over instances
   */
  for(int i = 0; i < nObject->NumInstances; i++) {  				
    wcstombs(name, (wchar_t *) InstanceName(nInstance), 80);										
    if( ( pid = ProcessGetPid(nObject, nInstance ) ) > 0 && !stricmp(appl, name) ) {
      *nThreads = ProcessGetNumThreads(gpPerfData, i);
      return pid;
    }
    
    // Get next instance
    nInstance = NextInstance(nInstance);		
  }
  
  return -1;
}

/**
 * Read in the registry values. Both key are read, NMapServArgs and 
 * NMapArgs. NMapServArgs contains the service parameter (startTime, 
 * interval, interval factor etc. NMapArgs the nmap parameters.
 */
bool CNMapService::ReadRegistry()
{
	int ret = 0;
	HKEY hkey;
	char szKey[1024];

	// Read the registry parameters. Try opening the registry key:
	// HKEY_LOCAL_MACHINE\SYSTEM\CurrentControlSet\Services\NMap\Parameters\NMapArgs
	strcpy( szKey, "SYSTEM\\CurrentControlSet\\Services\\NMap\\Parameters" );

	if( RegOpenKeyEx( HKEY_LOCAL_MACHINE, szKey, 0, KEY_QUERY_VALUE|KEY_NOTIFY, &hkey ) == ERROR_SUCCESS ) {

		// Yes we are installed
		DWORD dwType = REG_SZ;
		DWORD dwSize = sizeof(nmapArgs);

		// Get the nmap arguments
		if( RegQueryValueEx(hkey, "NMapArgs", NULL, &dwType, nmapArgs, &dwSize ) != ERROR_SUCCESS ) {

			logWriter->log( FATAL, "CNMapService", "ReadRegistry", 
							"Could not get args from registry" );
			*nmapArgs = '\0';
			return false;

		} else {

			if( dwSize <= sizeof( nmapArgs ) )
				logWriter->log( INFO, "CNMapService", "ReadRegistry", 
				                "Got nmap args: %s", nmapArgs );
			else {
				*nmapArgs = '\0';
				logWriter->log( INFO, "CNMapService", "ReadRegistry", 
				                "Args too long: %d", dwSize );
			}
		}

		// Get the nmapserv arguments
		// HKEY_LOCAL_MACHINE\SYSTEM\CurrentControlSet\Services\NMap\Parameters\NMapServArgs
		dwSize = sizeof( servArgs );
		dwType = REG_SZ;
		if( RegQueryValueEx(hkey, "NMapServArgs", NULL, &dwType, servArgs, &dwSize ) != ERROR_SUCCESS ) {

			logWriter->log( FATAL, "CNMapService", "ReadRegistry", 
							"Could not get service args from registry" );
			*servArgs = '\0';
			return false;

		} else {

			if( dwSize <= sizeof( servArgs ) )
				logWriter->log( INFO, "CNMapService", "ReadRegistry", 
								"Got service args: %s", servArgs );
			else
				*servArgs = '\0';
		}
		RegCloseKey(hkey);

	} else {

		logWriter->log( FATAL, "CNMapService", "ReadRegistry", "Could not open registry key" );
		return false;
	}
	return true;
}

/**
 * Parses the service arguments which are stored in the registry in the order:
 * hh:mm:ss: starttime
 * int:      interval
 * int:      factor
 * bool[7]:  days
 */
void CNMapService::ParseServiceArgs()
{
	char sTime[9];									// Temp time
	char* token = strtok( (char*)servArgs, " " );	// Token

	// Get the start time
	if( token )
		strcpy( sTime, token );
	if( token = strtok( NULL, " " ) )
		interval = atoi( token );
	if( token = strtok( NULL, " " ) )
		factor = atoi( token );

	// Get the active days (0=Sunday)
	for( int i = 0; i < 7; i++ ) {
		if( ( token = strtok( NULL, " " ) ) )
			days[i] = atob( token );
	}

	// Get the auto start flag
	if( ( token = strtok( NULL, " " ) ) )
		autoStart = atob( token );

	// Immediately flag
	if( ( token = strtok( NULL, " " ) ) )
		immediately = atob( token );

	// Split up the time
	token = strtok( sTime, ":" );
	if( token )
		hour = atoi( token );
	if( ( token = strtok( NULL, ":" ) ) )
		min = atoi( token );
	if( ( token = strtok( NULL, "\0" ) ) )
		sec = atoi( token );

	logWriter->log( DEBUG0, "CNMapService", "ParseServiceArgs", 
					"Got time: %d:%d:%d interval: %d factor: %d Days: %d %d %d %d %d %d %d AutoStart: %d Immediately: %d",
					hour, min, sec, interval, factor, days[0], days[1], days[2], 
					days[3], days[4], days[5], days[6], autoStart, immediately );
}

/**
 * Sets the next scan time.
 */
void CNMapService::GetNextScan()
{
	CTime now = CTime::GetCurrentTime();
	CTime next = CTime( now.GetYear(), now.GetMonth(), now.GetDay(), hour, min, sec );

	logWriter->log( DEBUG0, "CNMapService", "GetNextScan", "Factor: %d interval: %d", factor, interval); 

	while( next < now ) {
	
		// Calculate the next scan time
		switch( factor ) {

		case INTERVAL_MINUTES:
			next += interval * 60;
			break;

		case INTERVAL_HOURS:
			next += interval * 3600;
			break;

		case INTERVAL_DAYS:
			next += interval * 3600 * 24;
			break;

		case INTERVAL_WEEKS:
			next += interval * 3600 * 24 * 7;
			break;

		case INTERVAL_MONTHS:
			next += interval * 3600 * 24 * 30;
			break;
		}
	}

	logWriter->log( DEBUG1, "CNMapService", "GetNextScan", 
					"Got next scan time: %s", next.Format( "%d/%m/%y %H:%M:%S" ) );

	// Next scan time as UNIX time.
	nextScan = next.GetTime();
}

/**
 * Start the nmap network scanner as background process
 */
void CNMapService::StartScanner()
{
    int nThreads, pid;
    char appl[1024];

    logWriter->log( INFO, "CNMapService", "StartScanner", "Scanner starting" ); 

	if( strstr( (char*)nmapArgs, "127.0.0.1" ) ) {
	    logWriter->log( ERROR, "CNMapService", "StartScanner", "Cannot scan local loopback interface!" ); 
		return;
	}

    // If we have a process pid > 0
    if( ( pid = GetPid( "nmap", &nThreads ) ) < 0 ) {

		memset( &si, 0, sizeof( si ) ); 
		si.cb = sizeof(si); 

		sprintf( appl, "nmap.exe %s", nmapArgs );

		if( CreateProcess( NULL, appl, NULL, NULL, true, 
			DETACHED_PROCESS|CREATE_NO_WINDOW, NULL, NULL, &si, &pi ) )
			logWriter->log( INFO, "CNMapService", "StartScanner", "Started: %s", appl ); 
		else
			logWriter->log( ERROR, "CNMapService", "StartScanner", "Could not start: %s", appl ); 
	} else
		
		logWriter->log( ERROR, "CNMapService", "StartScanner", "NMap already running - pid: %d", pid ); 
}

/**
 * Main routine.
 */
int main(int argc, char *argv[])
{
	char* NMAPDIR;
	BOOL  bservice = true;
	BOOL  binstall = false;
	BOOL  buninstall = false;
	char  logFilename[256];
	int   debugLevel = 1;

	if( ( NMAPDIR = getenv( "NMAPDIR" ) ) == NULL ) {
		fprintf( stderr, "Environment variable NMAPDIR not set" );
		exit(1);
	}     

	if( signal( SIGINT, SignalHandler ) == SIG_ERR ) {
		fprintf(stderr, "Could not install signal handlers.");
		exit(1);
	}

	// Get the command line arguments 
	while((argc > 1) && (argv[1][0] == '-')) {

		switch(argv[1][1]) {
  
		case 'h':					
		  /* Print help screen and exit */
		  usage();
		  exit(0);
  
		case 'i':
		  // Start as background process
		  binstall = true;
		  break;

		case 'u':
		  // Start as background process
		  buninstall = true;
		  break;

		case 'p':
		  // Start Win32 service
		  bservice = false;
		  break;

		case 'd':
		  // Get the debug level.
		  ++argv;
		  --argc;
		  debugLevel = atoi( argv[1] );
		  break;

		default:
		  fprintf(stderr, "bad option %s\n", argv[1]);
		}
		++argv;
		--argc;
	}

	// Set default configuration and log file
	sprintf( logFilename, "%s\\log\\nmapserv.log", NMAPDIR );

	// Initialize the logging
	CLogWriter* logWriter;
	if( bservice )
		logWriter = new CLogWriter( logFilename, 4, 5, 5 );
	else
		logWriter = new CLogWriter( logFilename, debugLevel, 5, 5 );

	logWriter->log( INFO, "CNMapService", "InitLog", "Logging initialized" );
	CNMapService* nmapService = new CNMapService( logWriter );

	logWriter->log( INFO, "CNMapService", "Main", "CNMapService initialized" );

	if( binstall ) {

		logWriter->log( INFO, "CNMapService", "Main", "Trying to install service" );
	
		// Install the service
		if( nmapService->Install() ) {

			fprintf( stdout, "NMap Service installed." );
			exit(0);

		} else {

			fprintf( stdout, "Could not install NMap Service." );
			exit(1);

		}

	} else if( buninstall) {
    
		// Uninstall the service
		if( nmapService->Uninstall() ) {

			fprintf( stdout, "NMap Service uninstalled." );
			exit(0);

		} else {

			fprintf( stdout, "Could not uninstall NMap Service." );
			exit(1);

		}

	} else if( bservice ) {
    
	    logWriter->log( INFO, "CNMapService", "Main", "Starting service" );

		// Run as WIN32 service
		nmapService->StartService();

	} else {

		// Run as foreground process
		nmapService->m_bIsRunning = true;
		nmapService->OnInit();
		nmapService->RunService();
	}

	return(0);  
}

